
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

// hdyn_inline.C:  Inline functions for use in hdyn_ev.C, hdyn_harp3.C,
//		   and hdyn_slow.C.
//
//		   Include this file directly into the source files,
//		   rather than placing these functions in libraries.
//
// Local inline functions defined:
//
//	void update_nn_coll
//	real define_perturbation_radius_factor
//	bool is_perturber
//	real binary_scale
//
//----------------------------------------------------------------------

// inline void update_nn_coll
//	update the various neighbor pointers associated with a node.

local inline void update_nn_coll(hdyn* this_node, int id,
				 real distance_squared, hdyn * b,
				 real&d_nn_sq, hdyn*&nn,
				 real sum_of_radii,
				 real&d_coll_sq, hdyn*&coll)

// Note: arguments "this_node" and "id" are needed only for possible
// 	 diagnostic output.

{
    if (distance_squared < d_nn_sq) {
	d_nn_sq = distance_squared;
	nn = b;
    }

    real coll_dist = distance_squared - sum_of_radii*sum_of_radii;

    if (coll_dist < d_coll_sq) {
	d_coll_sq = coll_dist;
	coll = b;
    }
}


// Set/check criteria for placing bi on the perturber list of a node.
// Note that these functions are the *only* places in kira where this
// criterion is defined.

#define PERTURBER_CRITERION	2	// For determination of perturbers
					//   0:  distance only
					//   1:  perturbation (includes mass)
					//   2:  hybrid

// inline real define_perturbation_radius_factor()
// inline bool is_perturber()

// Simple distance criterion for placing bi on perturber list is
//
//	    distance_squared  <  perturbation_radius^2
//
// where
//
//	    perturbation_radius  =  gamma^{-1/3} * r_bin
//
// and
//
//	    r_bin  =  min(d_min, max(a, rij)).
//
// This is OK for roughly equal-mass binaries, but it doesn't take
// masses into account -- IMPORTANT when a wide mass range exists.
// (Also, it is not clear that d_min is the correct upper limit on
// the "size" of the binary r_bin.)
//
// The criterion including masses, which reduces to the above criterion
// for equal masses, is
//
//	      2 * m_pert * r_bin	       M_bin
//	    ---------------------  =  gamma * -------
//	    perturbation_radius^3	      r_bin^2
//
// or
//
//	    perturbation_radius  =  gamma^{-1/3} * r_bin
//	    				* (2 * m_pert / M_bin)^{1/3}.
//
// Note that, in the "perturbation" definitions, we work with cubed radii
// (to avoid cube roots in the loop that computes the perturber list).
// In these cases, "perturbation_radius_factor" is (apart from the mass
// factor) perturbation_radius^3.
//
// Also,in these cases, pert_factor_sq is really gamma2^{-1/3} = gamma^{-2/3}.
// All we really need is 1/gamma!  FIX THIS SOON!!
//
//*** Distance criterion is fine for perturbers of mass comparable to or
//*** less than the masses of the binary components.
//*** Hybrid criterion uses a fixed perturber radius, but then also includes
//*** more distant massive stars.

// define_perturbation_radius_factor:
//		 define the value of perturber_radius_squared.

local inline real define_perturbation_radius_factor(real pert_factor_sq,
						    real d_min_sq,
						    real r_bin,
						    real m_bin)
{
    if (PERTURBER_CRITERION == 0) {

	// Distance criterion:

	return pert_factor_sq * min(d_min_sq, r_bin * r_bin);

    } else if (PERTURBER_CRITERION == 1) {

	// Perturbation criterion (store rp^3 and include mass factor):

	real rp2 = pert_factor_sq * min(d_min_sq, r_bin * r_bin);
	return 2 * pow(rp2, 1.5) / m_bin;

    } else if (PERTURBER_CRITERION == 2) {

	// Hybrid criterion for perturbation radius is the distance
	// criterion, but storing rp^3 instead of rp^2; alternatively,
	// it is the perturbation criterion without the mass factor:

	real rp2 = pert_factor_sq * min(d_min_sq, r_bin * r_bin);
	return pow(rp2, 1.5);

    }
}

// binary_scale:  compute an appropriate "size" for a binary.

local inline real binary_scale(hdyn* cm)
{
    hdyn *d1 = cm->get_oldest_daughter();
    if (!d1) return 0;

    hdyn *d2 = d1->get_younger_sister();
    if (!d2) return 0;

    real sep = abs(d1->get_pos() - d2->get_pos());
    real energy = 0.5 * square(d1->get_vel() - d2->get_vel())
			- cm->get_mass() / sep;
    real sma = 0.5 * cm->get_mass() / abs(energy);

    return max(sma, sep);

    // return max(2*sma, sep);	// Better? -- more conservative, and also
				// avoids discontinuous changes in perturber
				// lists as the binary phase changes.
				// -- changes default gamma.
}

// is_perturber: check the perturbation criterion.

local inline bool is_perturber(real m_bin,		     // binary mass
			       real m_pert,		     // perturber mass
			       real distance_squared,
			       real perturbation_radius_factor)
{
    if (PERTURBER_CRITERION == 0) {

	// Distance criterion:

	return (distance_squared < perturbation_radius_factor);

    } else if (PERTURBER_CRITERION == 1) {

	// Perturbation criterion:

	return (distance_squared * sqrt(distance_squared)
					< m_pert * perturbation_radius_factor);

    } else if (PERTURBER_CRITERION == 2) {

	// Hybrid criterion is distance criterion, plus additional
	// massive perturbers:

	real d3 = distance_squared * sqrt(distance_squared);

	if (m_pert <= 0.5*m_bin)
	    return (d3 < perturbation_radius_factor);		// low-mass

	return (0.5*m_bin*d3					// high-mass
		     < m_pert * perturbation_radius_factor);

    }
}

// crit_separation_cubed: return the critical separation for the
//			  adopted perturbation criterion.

local inline real crit_separation_cubed(real m_bin,	// binary mass
					real m_pert,	// perturber mass
					real d_min_sq,	// global data
					real r_bin,	// binary scale
					real gamma)	// perturbation
{
    real d3 = pow(min(d_min_sq, r_bin * r_bin), 1.5) / gamma;

    if (PERTURBER_CRITERION == 0
	|| (PERTURBER_CRITERION == 2 && m_pert <= 0.5 * m_bin))

	return d3;

    else

	return 2 * (m_pert / m_bin) * d3;
}
