/*
 *  lagrad.c: Compute Lagrangian (mass) radii for an Nbody system
 *	      The radii are stored in the system (root) dyn story.
 *.............................................................................
 *    version 1:  May 1989   Piet Hut               email: piet@iassns.BITNET
 *                           Institute for Advanced Study, Princeton, NJ, USA
 *    version 2:  Dec 1992   Piet Hut  --  adopted to the new C++-based starlab
 *    version 3:  Jul 1996   Steve McMillan & Jun Makino
 *.............................................................................
 *  non-local functions: 
 *    compute_gereral_mass_radii
 *    compute_mass_radii_quartiles
 *    compute_mass_radii_percentiles
 *.............................................................................
 */
#include "dyn.h"

#ifndef TOOLBOX

#define SCRATCH_PAD_LINE_LENGTH 255
char scratch_pad[SCRATCH_PAD_LINE_LENGTH];

typedef  struct
{
    real  radius;
    real  mass;
} rm_pair, *rm_pair_ptr;

//-----------------------------------------------------------------------------
//  compare_radii  --  compare the radii of two particles
//-----------------------------------------------------------------------------

local int compare_radii(const void * pi, const void * pj)
{
    if (((rm_pair_ptr) pi)->radius > ((rm_pair_ptr) pj)->radius)
        return(1);
    else if (((rm_pair_ptr)pi)->radius < ((rm_pair_ptr)pj)->radius)
        return(-1);
    else
        return(0);
}

//-----------------------------------------------------------------------------
//  compute_general_mass_radii  --  Get the massradii for all particles.
//-----------------------------------------------------------------------------

void compute_general_mass_radii(dyn * b, int nzones)
{
    if (nzones < 2) return;

    real* mass_percent = new real[nzones-1];
    if (mass_percent == NULL) {
	cerr << "compute_general_mass_radii: "
	     << "not enough memory left for mass_percent\n";
	return;
    }

    int n = b->n_daughters();

    rm_pair_ptr rm_table = new rm_pair[n];
    if (rm_table == NULL) {
	cerr << "compute_general_mass_radii: "
	     << "not enough memory left for rm_table\n";
	return;
    }

    // Use the geometric center if the density center is unknown.

    vector dc_pos = 0;
    if (find_qmatch(b->get_dyn_story(), "density_center_pos"))
	dc_pos = getvq(b->get_dyn_story(), "density_center_pos");

    // Set up an array of (radius, mass) pairs.

    int i;
    dyn* bi;
    for (i = 0, bi = b->get_oldest_daughter(); bi != NULL;
	 i++, bi = bi->get_younger_sister()) {
	rm_table[i].radius = abs(bi->get_pos() - dc_pos);
	rm_table[i].mass = bi->get_mass();
    }	

    // Sort the array by radius.

    qsort((void *)rm_table, (size_t)n, sizeof(rm_pair), compare_radii);

    // Determine Lagrangian radii.

    for (i = 0; i < nzones-1; i++)
        mass_percent[i] = ((i + 1) / (real)nzones) * b->get_mass();

    real *rlagr = new real[nzones-1];
    real cumulative_mass = 0.0;
    i = 0;

    for (int k = 0; k < nzones-1; k++) {

        while (cumulative_mass < mass_percent[k])
	    cumulative_mass += rm_table[i++].mass;

	rlagr[k] = rm_table[i-1].radius;
    }

    // Place the data in the root dyn story.

    sprintf(scratch_pad, "rlagr[%d]", nzones-1);
    putra(b->get_dyn_story(), scratch_pad, rlagr, nzones-1);

}

// Convenient synonyms:

void  compute_mass_radii_quartiles(dyn * b)
{
    compute_general_mass_radii(b, 4);
}

void  compute_mass_radii_percentiles(dyn * b)
{
    compute_general_mass_radii(b, 10);
}

#else

/*-----------------------------------------------------------------------------
 *  main  --  driver to use  compute_mass_radii() as a tool
 *-----------------------------------------------------------------------------
 */

main(int argc, char ** argv)
{
    char  *comment;
    int n = 0;
    bool  c_flag = FALSE;      /* if TRUE: a comment given on command line   */
    bool  t_flag = FALSE;      /* if TRUE: percentiles rather than quartiles */

    extern char *poptarg;
    int c;
    char* param_string = "c:n:t";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c)
	    {
	    case 'c': c_flag = TRUE;
		      comment = poptarg;
		      break;
	    case 'n': n = atoi(poptarg);
		      break;
	    case 't': t_flag = TRUE;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      exit(1);
	    }            

    dyn *b;

    while (b = get_dyn(cin)) {

        if (c_flag == TRUE)
            b->log_comment(comment);

        b->log_history(argc, argv);

	if (t_flag)
	    compute_mass_radii_percentiles(b);
	else {
	    if (n <= 1)
		compute_mass_radii_quartiles(b);
	    else
		compute_general_mass_radii(b, n);
	}

	// Print out radii in case of quartiles only.

	if (find_qmatch(b->get_dyn_story(), "rlagr[3]")) {
	    real rlagr[3];
	    getra(b->get_dyn_story(), "rlagr[3]", rlagr, 3);
	    cerr << "rlagr =";
	    for (int i = 0; i < 3; i++) cerr << " " << rlagr[i];
	    cerr << endl;
	}

	put_dyn(cout, *b);	
	delete b;
    }
}

#endif

/* endof: lagrad.c */
