
//// mksecondary:  Create binary secondary components for randomly
////               selected stars in the input snapshot, placing the
////               results in a binary tree.  Only secondary masses
////               are set here; orbital parameters are set by
////               dyn::mkbinary.
////
//// Options:      -f    specify binary fraction [0.1]
////               -l    specify lower limit on mass ratio [1]
////               -s    specify random seed [random from system clock]

//		   Steve McMillan, July 1996

#include "node.h"

#ifdef TOOLBOX

// add_secondary: Add a secondary component to a primary to make
//		  a binary.  Only the mass, name and tree structure
//		  are modified at this stage.
//		  New primary and secondary components have names "na"
//		  and "nb", respectively, where "n" was the name of
//		  the original star.

local void add_secondary(node* original, real mass_ratio)
{
    node* primary = new node;
    node* secondary = new node;

    // Add new links.

    original->set_oldest_daughter(primary);

    primary->set_parent(original);
    secondary->set_parent(original);

    primary->set_younger_sister(secondary);
    secondary->set_elder_sister(primary);

    // Set new masses.

    primary->set_mass(original->get_mass());
    secondary->set_mass(mass_ratio*original->get_mass());
    original->inc_mass(secondary->get_mass());

    // Naming convention:

    if (original->get_name() == NULL)
	if (original->get_index() >= 0) {
	    char tmp[64];
	    sprintf(tmp, "%d", original->get_index());
	    original->set_name(tmp);
	}

    if (original->get_name() != NULL) {

	// Label components "a" and "b".

	primary->set_name(original->get_name());
	secondary->set_name(original->get_name());
	strcat(primary->get_name(), "a");
	strcat(secondary->get_name(), "b");

	// Make standard CM name.

	char tmp[256];
	sprintf(tmp, "(%s,%s)", primary->get_name(), secondary->get_name());
	original->set_name(tmp);
    }
//    cerr << "mksecondary: created binary (" << primary->get_name()
//	 << ", " << secondary->get_name() << ")\n";

}

local void mksecondary(node* b, real binary_fraction, real lower_limit)
{
    // For now, use a flat distribution in secondary mass ratio.
    // Assume that the probability of a star being the primary of
    // a binary is independent of mass.

    real sum = 0.5*binary_fraction;
    b->set_mass(0);

    for_all_daughters(node, b, bi) {
	sum += binary_fraction;
	if (sum >= 1) {
	    sum -= 1;

	    real mass_ratio = randinter(lower_limit, 1);	// Quick fix...
	    add_secondary(bi, mass_ratio);

	}
	b->inc_mass(bi->get_mass());
    }
}

void main(int argc, char ** argv)
{
    real binary_fraction = 0.1;
    real lower_limit = 1;
    int random_seed = 0;
    char seedlog[64];

    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "f:l:s:";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'f': binary_fraction = atof(poptarg);
		      break;
	    case 'l': lower_limit = atof(poptarg);
		      break;
	    case 's': random_seed = atoi(poptarg);
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      get_help();
		      exit(1);
	}

    if (binary_fraction < 0 || binary_fraction > 1)
	err_exit("mksecondary: Illegal binary fraction");
    if (lower_limit <= 0 || lower_limit > 1)
	err_exit("mksecondary: Illegal mass ratio limit");

    node* b;
    b = get_node(cin);

    b->log_history(argc, argv);

    int actual_seed = srandinter(random_seed);

    sprintf(seedlog, "         random number generator seed = %d",actual_seed);
    b->log_comment(seedlog);

    mksecondary(b, binary_fraction, lower_limit);

    put_node(cout, *b);
    rmtree(b);
}
#endif
