//
// double_star.C
//

#include "double_star.h"

double_star * new_double_star(node* n, real sma, real ecc, 
			      real start_time=0, int id=0,
                              binary_type type = Detached) {

     double_star* element = new double_star(n);
     element->initialize(type, sma, ecc, start_time, id);

     element->get_primary()->set_current_time(start_time);
     element->get_secondary()->set_current_time(start_time);

     element->refresh_memory();       //implemented 25-1-95
     element->set_donor_timescale(element->get_primary());
     element->determine_minimal_timestep();
/*set_minimal_timestep(
              element->get_donor_timescale()*MIN_DT_FACTOR);
*/

     return element;
}

double_star::double_star(node* n) : star(n) {

           semi=eccentricity=binary_age=minimal_timestep=velocity=0;
           donor_timescale=0;
           identity = donor_identity=0;
           donor_type=NAS;
           first_contact=FALSE;
        }

void double_star::initialize(binary_type type, real sma, 
  			real ecc, real age, int id) {
   
      semi = sma;
      eccentricity = ecc;
      bin_type = type;
      binary_age = age;
      identity = id;

      initial.semi = semi;
      initial.q = mass_ratio();
      initial.mass_prim = get_primary()->get_total_mass();
      initial.eccentricity = eccentricity;

      refresh_memory();       //implemented 25-1-95
      set_donor_timescale(get_primary());
      minimal_timestep = MIN_DT_FACTOR*donor_timescale;
   }

real double_star::mass_ratio()           {
     real mp = get_primary()->get_total_mass();
     real ms = get_secondary()->get_total_mass();

     real q = ms/mp;
     if (q>1) q = mp/ms;

     return q;
  }

real double_star::roche_radius(star * str) {
//             Eggleton PP., ApJ, 1983, 268, 368.
               real mr = str->get_total_mass()
                       / str->get_companion()->get_total_mass();
               real q1_3 = pow(mr, ONE_THIRD);
               real q2_3 = q1_3*q1_3;   //pow(mr, TWO_THIRD);
               return semi*0.49*q2_3/(0.6*q2_3 + log(1 + q1_3));
        }
real double_star::roche_radius(const real a, const real m1, const real m2) {
               real q = m1/m2;
               real q1_3 = pow(q, ONE_THIRD);
               real q2_3 = q1_3*q1_3;   //pow(mr, TWO_THIRD);
               return a*0.49*q2_3/(0.6*q2_3 + log(1 + q1_3));
       }

void double_star::put_element() {
//              should be made starlab compatible

     ofstream outfile("binev.dat", ios::app|ios::out);
     if(!outfile) cerr << "error: couldn't crate file binev.dat"<<endl;

//     swap_elements();

          outfile << identity << "\t" << binary_age << endl;
          outfile << identity << endl;
          get_primary()->put_element();
          get_secondary()->put_element();

     initial.put_element();
     outfile << bin_type << " " << eccentricity << " "
             << get_period() << " " << semi << " " << velocity << "\n" << endl;

     outfile.close();
        }

void double_star::put_hrd(ostream & s) {

//        swap_elements();
        s << "2\n  ";
        get_primary()->put_hrd(s);
        s << "  "; 
        get_secondary()->put_hrd(s);
   }

//	Pretty print.
void double_star::print_status() {
        cout << "status at time = " << binary_age << endl;

        if (get_primary()->no_of_elements()>1) 
           get_primary()->print_status();
        if (get_secondary()->no_of_elements()>1)   
           get_secondary()->print_status();

        cout << type_string(bin_type);
        switch (bin_type) {
           case Merged: 
              cout << " {" << type_string(get_primary()->get_element_type())
                   << "}" << endl;
              cout << "M = "<<get_primary()->get_total_mass()
                   << "\t R = "<<get_primary()->get_effective_radius()
                   << "\t V = "<<get_primary()->get_velocity() << endl;
              break;
           case Disrupted:
              cout << " )" << type_string(get_primary()->get_element_type())
                   << ", " << type_string(get_secondary()->get_element_type())
                   << "(" << endl;
              cout << "M = "<<get_primary()->get_total_mass()
                   << "\t R = "<<get_primary()->get_effective_radius()
                   << "\t V = "<<get_primary()->get_velocity() << endl;
              cout << "m = "<<get_secondary()->get_total_mass()
                   << "\t r = "<<get_secondary()->get_effective_radius()
                   << "\t v = "<<get_secondary()->get_velocity() << endl;
              break;
           default:
              if (get_primary()->get_spec_type(Rl_filling))
                 cout << " [";
              else
                 cout << " (";
              cout << type_string(get_primary()->get_element_type())
                   << ", " << type_string(get_secondary()->get_element_type());
              if (get_secondary()->get_spec_type(Rl_filling))
                 cout << "]" << endl;
              else
                 cout << ")" << endl;
              cout << "M = "<<get_primary()->get_total_mass()
                   << "\t R = "<<get_primary()->get_effective_radius() << endl;
              cout << "m = "<<get_secondary()->get_total_mass()
                   << "\t r = "<<get_secondary()->get_effective_radius() << endl;
              cout << "a = "<<semi<<"\t e = "<<eccentricity
                   << "\t v = "<<get_velocity() << endl;
           }
     }

//	Pretty-print
void double_star::put_state() {
       
//        swap_elements();
        star *a, *b;
        if(get_primary()->get_identity()<get_secondary()->get_identity()) {
          a = get_primary();
          b = get_secondary();
        }
        else {
          b = get_primary();
          a = get_secondary();
        }
        switch (bin_type) {
           case Merged:
              cout << " {"; 
              get_primary()->put_state();
              cout << "}" << endl;
              break;
           case Disrupted:
              cout << " )";
              a->put_state();
              cout << ", "; 
              b->put_state();
              cout << "(" << endl;
              break;
           default:
              if (a->get_spec_type(Rl_filling))
                 cout << " [";
              else
                 cout << " (";
              a->put_state();
              cout << ", ";
              b->put_state();
             if (b->get_spec_type(Rl_filling))
                 cout << "]" << ":p=" << get_period() << endl;
              else
                 cout << ")" << ":p=" << get_period() << endl;
           }
     }

void double_star::print_roche() {
//	Gives compatible output for EPJ Roche program.

     if (bin_type!=Merged) {
        if (get_primary()->get_identity()<get_secondary()->get_identity()) {
           get_primary()->print_roche();
           get_secondary()->print_roche();
        }
        else {
           get_secondary()->print_roche();
           get_primary()->print_roche();
        }
     }
     else get_primary()->print_roche();

     cout<<get_period();
   }

void double_star::dump(ostream & s) {
//	Zero IQ dump program.

      if(bin_type!=Merged) {
//      swap_elements();
      int n_elements = no_of_elements();

      s << n_elements
        << "\n " << identity
        << " " << binary_age
        << " " << bin_type
        << " " << eccentricity
        << " " << get_period()
        << " " << semi
        << " " << velocity
        << endl;
      initial.dump(s);

      s << "  ";
      get_primary()->dump(s);
      s << "  ";
      get_secondary()->dump(s);
      }
      else get_primary()->dump(s);

   }

void double_star::dump(char * filename) {

      ofstream s(filename, ios::app|ios::out);
      if(!s) cerr << "error: couldn't crate file " << filename <<endl;

//      swap_elements();
      int n_elements = no_of_elements();

      s << n_elements
        << "\n " << identity
        << " " << binary_age
        << " " << bin_type
        << " " << eccentricity
        << " " << get_period()
        << " " << semi
        << " " << velocity
        << endl;
      initial.dump(s);

      s << "  ";
      get_primary()->dump(s);
      s << "  ";
      get_secondary()->dump(s);

      s.close();
   }

void double_star::adjust_binary_after_wind_loss(star * donor,
                                                const real md_dot,
                                                const real dt) {
//	Binary separation is adjusterd upon mass loss by
//	stellar wind of one of the companions. 
//	Fraction of the mass lost by the donor is dumped on its
//	companion.


//		check if object is still bound.
     if (bin_type!=Merged && bin_type!=Disrupted) {
        star* accretor = donor->get_companion();

        real M_old = get_total_mass();
        real old_donor_mass = donor->get_total_mass();
        real old_accretor_mass = accretor->get_total_mass();

        real ma_dot = accretor->accrete_from_stellar_wind(md_dot, dt);
        donor = donor->reduce_donor_mass(md_dot);

        real M_new = get_total_mass();
        real new_donor_mass = donor->get_total_mass();
        real new_accretor_mass = accretor->get_total_mass();

        if (md_dot>0 && ma_dot>=0) {
           real alpha = 1 - ma_dot/md_dot;
           semi *= pow(pow(new_donor_mass/old_donor_mass, 1-alpha)
                *  new_accretor_mass/old_accretor_mass, -2)
                *  M_old/M_new;
        }
     }
     else {
//		Not a binary: reduce only the mass of the donor.
//        donor->reduce_donor_mass(md_dot);
        //try {
        donor = donor->reduce_donor_mass(md_dot);
        //}
        //catch (star* element) {
         //  donor = element;
        //}
     }

//cerr<<" a-->"<<semi<<endl;
}

real double_star::angular_momentum() {
//	Determine the angular momentum of the binary.

// 	Shore, S.N., Livio, M., Heuvel, EPJ, 1992, in Interacting Binaries
// 	(edt. N Nussbaumer and A. Orr) Spirnger verlag p.\ 389.
        real omega = TWO_PI/(get_period()*DAYS);
        real mp = get_primary()->get_total_mass()*M_SUN;
        real ms = get_secondary()->get_total_mass()*M_SUN;
        real a =  semi*R_SUN;

        return omega*mp*ms*a*a*(1-eccentricity*eccentricity)/(mp+ms);
     }

void double_star::circularize() {
//	Tidal circularization with 
//	conservation of angular momentum.
//	Binary is not nesecerely circularized but 
//	periostron adapted to current TIDAL_RANGE.
//	This method is more realistic than instantanious 
//	curcularization.

     real pericenter = semi*(1-eccentricity);

     if((bin_type != Merged && bin_type != Disrupted)          &&
        (pericenter<=TIDAL_RANGE*get_primary()->get_radius()         ||
         pericenter<=TIDAL_RANGE*get_secondary()->get_radius())      &&
        (eccentricity>0 && eccentricity<1.)) /*safety*/         {
            real peri_new = TIDAL_RANGE*max(get_primary()->get_radius(),
				       get_secondary()->get_radius());
            real circ_semi = semi*(1-eccentricity*eccentricity);
            real new_ecc = circ_semi/peri_new - 1;
//cerr<<"old: "<<" "<<semi<<" "<<eccentricity<<" "<<pericenter<<endl;
//cerr<<"new: "<<circ_semi<<" "<<new_ecc<<" "<<peri_new<<endl;
            if(new_ecc<=0) {
              semi=circ_semi;
              eccentricity = 0;
            }
            else {
              semi = circ_semi/(1 - new_ecc*new_ecc);
              eccentricity = new_ecc;
            }
//cerr<<"final: "<<semi<<" "<<eccentricity<<endl;
            //semi *= (1-eccentricity*eccentricity);
            //eccentricity = 0;
            real to_km_per_s = 2*PI*R_SUN/(DAYS*Km_s);
            get_primary()->set_rotational_velocity(
                     to_km_per_s*get_primary()->get_radius()/get_period());
            get_secondary()->set_rotational_velocity(
                     to_km_per_s*get_secondary()->get_radius()/get_period());
        }
//force_circularization();	//	For model E.
}

void double_star::force_circularization() {
//	Implemented for model E.
//	Makes binaries immediately circular.

     if((bin_type != Merged && bin_type != Disrupted) &&
        (eccentricity>0 && eccentricity<1.)) /*safety*/          {
            semi *= (1-eccentricity*eccentricity);
            eccentricity = 0;
            real to_km_per_s = 2*PI*R_SUN/(DAYS*Km_s);
            get_primary()->set_rotational_velocity(
                     to_km_per_s*get_primary()->get_radius()/get_period());
            get_secondary()->set_rotational_velocity(
                     to_km_per_s*get_secondary()->get_radius()/get_period());
        }
}

void double_star::determine_minimal_timestep() {
//	Minimum timestep determination is needed for
//	preventing infinit integration upon dynamical timescale
//	mass-transfer.

     minimal_timestep = MIN_DT_FACTOR*donor_timescale;

     if(minimal_timestep<ABSOLUTE_DT_MIN)
        minimal_timestep = ABSOLUTE_DT_MIN;
}

real double_star::determine_dt(const real ageint, const real time_done) {
//	Determined the evolution timestep.
//	This is the hard of the program.
//	DO NOT CHANGE THIS ROUTINE!

     real dtime = ageint - time_done;
        real dtime1 = get_primary()->get_evolve_timestep();
        real dtime2 = Maximal_timestep;
        if (bin_type != Merged)
           dtime2 = get_secondary()->get_evolve_timestep();
        if(dtime1>dtime2) dtime1=dtime2; 
        if(dtime>dtime1) dtime = dtime1;
        if(dtime>minimal_timestep && bin_type==Semi_Detached) 
           dtime = minimal_timestep;

//		Leaving this out might be interesting.
//		At least it provides an enourmous speed increase
//		for low mass systems.
     if(bin_type != Merged && bin_type != Disrupted) {
        if(dtime>Maximal_timestep) dtime = Maximal_timestep; // Safety!!!
    }

     if(dtime<ABSOLUTE_DT_MIN) dtime=ABSOLUTE_DT_MIN;

     return dtime;
}

void double_star::force_donor_timescale(star* donor) {

     donor_identity = donor->get_identity();
     donor_type = donor->get_element_type();
     donor_timescale = donor->mass_transfer_timescale();
}

void double_star::binary_in_contact(const real dt) {
//	Search which star is actually filling its Roche lobe.

     real rl_p = roche_radius(get_primary());
     real rl_s = roche_radius(get_secondary());

     if (get_primary()->get_radius() >= rl_p) {
        if (get_secondary()->get_radius() < rl_s) {
           if (!ready_for_mass_transfer(get_primary())) return;
           semi_detached(get_primary(), get_secondary(), dt);
        }
        else
             common_envelope(dt);
     }
     else if (get_secondary()->get_radius() >= rl_s) {
        if (!ready_for_mass_transfer(get_secondary())) return;
        semi_detached(get_secondary(), get_primary(), dt);
     }
     cerr<<"leave: binary_in_contact"<<endl;
}

bool double_star::ready_for_mass_transfer(star* donor) {
//	Make sure that the star that fills its Roche lobe is the same as
//	the one that provided its mass-transfer timescale.

     real old_dt_min = minimal_timestep;
     bool mass_transfer_allowed = TRUE;

     if (donor->get_identity()!=donor_identity) {
        donor_identity    = donor->get_identity();
        donor_type        = donor->get_element_type();
        donor_timescale   = donor->mass_transfer_timescale();

        determine_minimal_timestep();

        if(old_dt_min<=minimal_timestep) 
           mass_transfer_allowed = TRUE;
        else
           mass_transfer_allowed = FALSE;
     }

     return mass_transfer_allowed;
}

       

void double_star::semi_detached(star* donor,
                                star* accretor,
                                real dt) {
//	Binary is semi-detached.
//	Donor should lose mass while the accretor should accrete.
    cerr << "semi_detached" <<endl;
    PRL(donor_timescale);
     if (!stable()) {
        if (donor->get_radius()>accretor->get_radius())
           spiral_in(donor, accretor);
        else
           spiral_in(accretor, donor);
     }
     else 
        perform_mass_transfer(dt, donor, accretor);

    cerr<<"prepare to update_binary parameters:..."<<endl;
     if (bin_type!=Merged && bin_type!=Disrupted)
        update_binary_parameters();
cerr<<"leave: double_star::semi_detached donor: "<<donor<<" accretor: "
    <<accretor<<endl;
     }

void double_star::contact(star* donor,
                          star* accretor,
                          real dt) {
//	Both stars in contact.
//	W Uma (stable) system or spiral-in may occur.

    cerr<<"double_star::contact()"<<endl;
//      Mass loss according to AM accretor.
     real M_old = get_total_mass();
     real old_donor_mass = donor->get_total_mass();
     real old_accretor_mass = accretor->get_total_mass();
     real q_old = old_accretor_mass/old_donor_mass;

//	Old mechanism
//     real md_dot = 0;
//     md_dot = donor->subtrac_mass_from_donor(dt, md_dot);

//	New mechanism
     real md_dot=0;
     donor = donor->subtrac_mass_from_donor(dt, md_dot);
/*
 *    try { 
 *       donor->subtrac_mass_from_donor(dt, md_dot);
 *      }
 *    catch (star * element) {
 *       donor = element;
 *    }
 */
     if (md_dot>0) {
        real ma_dot = accretor->add_mass_to_accretor(md_dot, dt);

        real M_new = get_total_mass();
        real new_donor_mass = donor->get_total_mass();
        real new_accretor_mass = accretor->get_total_mass();
        real q_new = new_accretor_mass/new_donor_mass;

        real a_fr;
        if (!accretor->remnant()) {
           real beta = 3;
           a_fr  = pow(old_donor_mass*old_accretor_mass
                 / (new_donor_mass*new_accretor_mass), 2);
           semi *= pow(M_new/M_old, 2*beta + 1)*a_fr;
        }
        else {
           real alpha = 1 - ma_dot/md_dot;
           if (alpha<1) {
              a_fr  = (new_donor_mass/old_donor_mass)
                    * pow(new_accretor_mass/old_accretor_mass, 1/(1-alpha));
              semi *= (M_old/M_new)/pow(a_fr, 2);
           }
           else {
              a_fr  = exp(2*(new_donor_mass-old_donor_mass)/new_accretor_mass);
              semi *= (M_old/M_new)*a_fr/pow(new_donor_mass/old_donor_mass, 2);
           }
        }

       real t_donor = donor->mass_transfer_timescale();
       if (((q_old<=1 && q_new>=1) ||
          (t_donor<0.1*donor_timescale || t_donor>10*donor_timescale)) &&
          (bin_type!=Disrupted && bin_type!=Merged)) {
          set_donor_timescale(t_donor,
                              donor->get_element_type(),
                              donor->get_identity());
          determine_minimal_timestep();
       }
     }

     if (bin_type!=Merged && bin_type!=Disrupted)
        update_binary_parameters();
}

bool  double_star::stable() {
//	Stability test for mass transfer.
//	Single stellar angular momentum should be smaller than
//	1/3-th the binary angular momentum.
      
        real J_prim = get_primary()->angular_momentum();
        real J_sec  = get_secondary()->angular_momentum();
        real J_star = max(J_prim, J_sec);
        real J_bin  = angular_momentum(); // + J_prim + J_sec;
cerr<<"I="<<J_prim<<" "<<J_sec<<" "<<J_bin<<endl;

//		For (ms, ms) binaries it does not matter too much
//		wheather or not mass-transfer is stable
//		but mass transfer to a remnant might be less 
//		stable.
        real factor = 0.7;
        if ((get_primary()->remnant() && !get_secondary()->remnant()) ||
            (!get_primary()->remnant() && get_secondary()->remnant())) 
            factor = 0.3;
        else if(get_primary()->remnant() && get_secondary()->remnant())
            factor = 3.0;

//		In principle this means the following:
//		0.42 J* >= Jb/3.	(st, st): -instable-
//		0.89 J* >= Jb/3.	(st, r):  -very instable-
//		0.09 J* >= Jb/3.	(r, r):	  -stable-
//		Here we assumed for symplicity k^2 to be unity (k^2=1).
//		In reality: 0.01 < k^2 < 0.4
//		k^2=0.4 mean a homogeneous globe.
//		Note that mass transfer between two remnants generally
//		leads to a merger anyway.
//		In the case of a [he, ns) the helium star is already in
//		corotation with the binary orbit while the neutron star 
//		will not be forced in corotation due to its small size.
//		This can stabilize the binary orbit dramatically.

        if (J_star>=factor*J_SPI_FR*J_bin) 
           return FALSE;
 
        return TRUE;
}

void double_star::update_binary_parameters() {
//	Update the evolved binary.

cerr<<"enter: double_star::update_binary_parameters()"<<endl;
//     swap_elements();

     real primary_mass = get_primary()->get_total_mass();
     real secondary_mass = get_secondary()->get_total_mass();

     get_primary()->set_previous_radius(get_primary()->get_radius());
     get_secondary()->set_previous_radius(get_secondary()->get_radius());
cerr<<"leave: double_star::update_binary_parameters()"<<endl;
}


void double_star::perform_mass_transfer(const real dt,
                                   star * donor,
                                   star * accretor) {
//	Adjust binary parameter according to the model
//	Take account of mass lost and transfered from the donor to the
//	accretor.


cerr<<"enter perform_mass_transfer("<<dt<<", "
    <<donor->get_element_type()<<", "<<accretor->get_element_type()
    <<") semi= "<< semi<<endl;

//      Mass loss according to AM accretor.
     real M_old = get_total_mass();
     real old_donor_mass = donor->get_total_mass();
     real old_accretor_mass = accretor->get_total_mass();
     real q_old = old_accretor_mass/old_donor_mass;
//cerr<<"m: "<< M_old<<" " << old_donor_mass <<" : " 
//           <<old_accretor_mass << " " << q_old << endl;

//     real md_dot=0;
//     md_dot = donor->subtrac_mass_from_donor(dt);
//		Reducing mass from the donor gave problems in the past.
//		The class star is absolutely needed in order to prevent
//		address changes during mass reduction procedures.
//		The previous solution was:
//		Do not use it anymore.
     real md_dot=0;
     donor = donor->subtrac_mass_from_donor(dt, md_dot);

//		Still an older version
//     star * old_d = donor;
//     md_dot = donor->subtrac_mass_from_donor(dt);
//     if (old_d==primary) donor = primary;
//     else donor = secondary;
//cerr<<"md_dot : " << md_dot<<endl;

     if (md_dot>0) {
        real ma_dot = accretor->add_mass_to_accretor(md_dot, dt);

        real M_new = get_total_mass();
        real new_donor_mass = donor->get_total_mass();
        real new_accretor_mass = accretor->get_total_mass();
//cerr<<"ma_dot>0 :"<<ma_dot<<" " <<M_new<<" " <<new_donor_mass<<" "
//   << new_accretor_mass<<endl;
        real q_new = new_accretor_mass/new_donor_mass;
//cerr<<"md_dot>0 :"<<M_new<<" " <<new_donor_mass<<" "
//   << new_accretor_mass<<" " << q_new<<endl;

        real a_fr;
        if (!accretor->remnant()) {
           real beta = 3;	// variable beta should be made global or
				// mass-transfer conservation depended.
           a_fr  = pow(old_donor_mass*old_accretor_mass
                 / (new_donor_mass*new_accretor_mass), 2);
           semi *= pow(M_new/M_old, 2*beta + 1)*a_fr;
        }
        else {
           real alpha = 1 - ma_dot/md_dot;		// alpha is a bad choice.
           if (alpha<1) {
              a_fr  = (new_donor_mass/old_donor_mass)
                    * pow(new_accretor_mass/old_accretor_mass, 1/(1-alpha));
              semi *= (M_old/M_new)/pow(a_fr, 2); 
           }
           else {
              a_fr  = exp(2*(new_donor_mass-old_donor_mass)/new_accretor_mass); 
              semi *= (M_old/M_new)*a_fr/pow(new_donor_mass/old_donor_mass, 2);
           }
        }

//cerr<<"new_semi: "<< a_fr<<" " <<semi<<endl;
//		Let outer component accrete from mass lost of inner binary.
//	Routine block for triple evolution.
//	Tricky but fun!
//        if (is_binary_component()) {
        if (is_star_in_binary()) {
           real mdot_triple = M_old - M_new;
           if (mdot_triple>0) {
           get_binary()->adjust_triple_after_wind_loss(this, mdot_triple, dt);
        }
        else if (is_binary_component()) {
           if(mdot_triple<0) {
           cerr << "enter perform_mass_transfer(" << dt << ", "
                << donor->get_element_type()<<", "
                <<accretor->get_element_type()
                <<")"<<endl;
           cerr<<"Mass lost during non-conservative mass transfer is negative."
               <<"mlost ="<<mdot_triple<<endl;
           }
           }
	else {
              cerr<<"Presumed binary component is not a binary root, "
                  << "nor a hierarchical root"<<endl;
           }
        }

//	Adjust mass-transfer timescale if needed.
        real t_donor = donor->mass_transfer_timescale();
        if (((q_old<=1 && q_new>=1) ||
           (t_donor<0.1*donor_timescale || t_donor>10*donor_timescale)) && 
           (bin_type!=Disrupted && bin_type!=Merged)) {
              set_donor_timescale(t_donor, 
                           donor->get_element_type(), 
                           donor->get_identity());
              determine_minimal_timestep();
        }
    }
cerr<<"leave : double_star::perform_mass_transfer() "<<endl;
cerr<<"pointer structure: " << endl;
cerr<<"this: " << this<< " ibc?: " << is_binary_component()
     <<" isib? " << is_star_in_binary() << " ib:? " <<  is_binary()
     << endl;
cerr<<"donor: "<<donor<<" "<<donor->get_companion()<<" "<<
    donor->is_binary_component()<<" "<<donor->is_star_in_binary()
    <<" "<<donor->is_binary()<<endl;
cerr<<"accretor: "<<accretor<<" "<<accretor->get_companion()<<" "<<
    accretor->is_binary_component()<<" "<<accretor->is_star_in_binary()
    <<" "<<accretor->is_binary()<<endl;
dump(cerr);
print_status();
}

void double_star::post_sn_companion_velocity(star* companion,
                                        const real mdot) {
//		Result of impact on companion velocity!
//		Temorary removed!
        return ;

//		Statement never reached.
//		Removed 16_01_94!
//cerr<<"void double_star::post_sn_companion_velocity()"<<endl;
//cerr<<"pre sn velocity: "<< companion->get_velocity();
     real new_total_binary_mass = get_total_mass()-mdot;
     real vel = abs((1 - 2*new_total_binary_mass/get_total_mass())
                           *pow(get_total_mass()
                           /companion->get_total_mass(), 2));
     //if(vel<=-1) vel = -1;
     vel = companion->get_velocity()*sqrt(1 + vel);
     companion->set_velocity(vel);
//cerr<<"post sn velocity: "<< companion->get_velocity()<<endl;
}

void double_star::evolve_element(const real end_time) {

     real current_time = binary_age
                       + BINARY_UPDATE_TIMESTEP*get_evolve_timestep();
 cerr<<"pre_ev: " <<binary_age<<" " << current_time<<" " << end_time
     << endl;
     print_status();
     if (current_time<=end_time)
	 do {
	     evolve_the_binary(current_time);
	     current_time += BINARY_UPDATE_TIMESTEP*get_evolve_timestep();
	 }
	 while (current_time<=end_time);
 cerr<<"post: " << binary_age<<" " <<current_time<<" " << end_time << endl;
     }

void double_star::evolve_the_binary(const real end_time) {
//	Evolve both stars synchonously.

     real dt, old_binary_age;
     real time_done = 0;
     real ageint = end_time - binary_age;

     do {
	 cerr<<"do a loop"<<endl;
        determine_minimal_timestep();
        dt = determine_dt(ageint, time_done);
        recursive_counter = 0; // Just checking!!, do not use!
        old_binary_age = binary_age;
        get_primary()->set_previous_radius(get_primary()->get_radius());
        get_secondary()->set_previous_radius(get_secondary()->get_radius());
        perform_wind_loss(dt);
         angular_momentum_loss(dt);

        refresh_memory(); 
        recursive_binary_evolution(dt, binary_age+dt);
	cerr<<"returned from recursve call"<<endl;
        calculate_velocities();
	cerr<<"velocities computed"<<endl;
        time_done += binary_age-old_binary_age; //dt; //!!! 21-0793!! TRICKY;
cerr<<"do another loop if: "<< time_done<<"<"<<ageint<<endl;
     }
     while(time_done<ageint);
}

void double_star::recursive_binary_evolution(real dt,
                                const real end_time) {
//	Evolve both stars synchronously.
//	In one star start filling its Roche-lobe
//	return one time-step and search for the exact moment
//	of Roche lobe contact.
//	then star to transfer mass.
//	Search procedure is performed with lables put onto the
//	double_star.
//	The procedure is tricky, sensitive to the timestep,
//	but robust.

//cerr<<" -r- ";
//cerr << "recursive_binary_evolution(dt " << dt 
//     << ", end_time " << end_time << "): " <<bin_type<< endl;
//cerr<<dt<<" "<<binary_age<<" "<<end_time<<endl;

//	More a ckeck than really needed.
//	Prevents number of recursive calls to grow infinitly.
      recursive_counter++;
    if(recursive_counter>=1000)
       return;

     if (binary_age>=end_time) return;     //SAFETY!!!!
     binary_age += dt;

//      Increase stellar ages and update parameters.
     if (get_primary())   get_primary()->evolve_element(binary_age);
     if (bin_type!=Merged) {
        if(get_secondary()) get_secondary()->evolve_element(binary_age);
     }

//	Are we still talking about a binary?
    if (bin_type!=Merged && bin_type!=Disrupted) {

       circularize();

//      Determine primary and secondary.
//       swap_elements();

       real primary_mass = get_primary()->get_total_mass();
       real secondary_mass = get_secondary()->get_total_mass();
       real rl_p = roche_radius(get_primary());
       real rl_s = roche_radius(get_secondary());

//		Just for memory and output routines.
//		Primary fills Roche-lobe.
       if (get_primary()->get_radius() >= rl_p) {
          get_primary()->set_effective_radius(rl_p);
          get_primary()->set_spec_type(Rl_filling);
          get_secondary()->set_spec_type(Accreting);
//          if (circularization_radius(secondary_mass, primary_mass)
//              >=secondary->get_radius()) {
//             dump("stream_hit.dat");  
//          }
          if (!first_contact) {
             first_contact=TRUE;
             first_roche_lobe_contact_story(
	        get_primary()->get_element_type(),
	        get_secondary()->get_element_type());
          }
       }

//		Secondary fills Roche-lobe.
       if (get_secondary()->get_radius() >= rl_s) {
          get_secondary()->set_effective_radius(rl_s);
          get_secondary()->set_spec_type(Rl_filling);
          get_primary()->set_spec_type(Accreting);
//          if (circularization_radius(primary_mass, secondary_mass)
//              >=primary->get_radius()) {
//             dump("stream_hit.dat");
//          }
          if (!first_contact) {
             first_contact=TRUE;
             first_roche_lobe_contact_story(
                get_secondary()->get_element_type(),
                get_primary()->get_element_type());
          }
       }

//		Determines if we really have to do with a mass
//		transferring binary.
       if (get_primary()->get_radius() >= rl_p    ||
           get_secondary()->get_radius() >= rl_s) {
          if ((get_primary()->get_radius() >= rl_p    &&
               get_secondary()->get_radius() >= rl_s) ||
              bin_type==Contact) {
             bin_type = Contact;
             common_envelope(dt);
             refresh_memory();
             if (binary_age>=end_time) return;     //SAFETY!!!!
             else
                dt = end_time - binary_age;
           }
           else if(dt<=minimal_timestep) {
              binary_in_contact(dt);
              refresh_memory();
cerr<<"Return if: "<<binary_age<<">="<<end_time<<endl;
              if (binary_age>=end_time) return;     //SAFETY!!!!
              dt = minimal_timestep;  //03_02_94 !! SAFETY!!
cerr<<"Minimal timestep is set"<<endl;
              if (binary_age + dt>end_time)
                 dt = end_time - binary_age;
cerr<<"Go for a new recursive call?"<<endl;
              if (bin_type!=Merged && bin_type != Disrupted) {
                 recursive_binary_evolution(dt, end_time);
              }
           }
           else {
              bin_type = Semi_Detached;
//     		Total recall!!!
              recall_memory();
              dt *= 0.5;
              recursive_binary_evolution(dt, end_time);
           }
        }
        else if(bin_type == Semi_Detached) {
           bin_type = Detached;
           if(dt>minimal_timestep)  //CAREFULL!! {} implemented 3_2_94
//              Refresh Memory!!!
           refresh_memory();
           dt *= 0.5;
           recursive_binary_evolution(dt, end_time);
        }
        else {
           bin_type = Detached;
           real maximal_timestep = end_time - binary_age;
           if (dt<maximal_timestep) {
//              Refresh Memory!!!
              refresh_memory();
              dt = maximal_timestep;
              recursive_binary_evolution(dt, end_time);
           }
        }
     }

     real max_dt = end_time - binary_age;
     if (dt-max_dt<max_dt) {
        if (bin_type!=Merged && bin_type!=Disrupted)
           bin_type = Detached;

//              Refresh Memory!!!
        refresh_memory();
        dt = max_dt;
        recursive_binary_evolution(dt, end_time);
     }
      cerr<<"leave: recursive_binary_evolution"<<endl;
}

void double_star::set_donor_timescale(star * donor) {

cerr <<"void double_star::set_donor_timescale(donor="
     << donor->get_element_type()<<")"<<endl;

     if (bin_type!=Merged && bin_type!=Disrupted) {
        int id          = donor->get_identity();
        stellar_type st = donor->get_element_type();
        real t          = donor->mass_transfer_timescale();

        if (st!=NAS) {
           if (donor_identity!=id) {
              donor_identity = id;
              donor_type = st;
              donor_timescale = t;
           }
           else if (donor_timescale<=0 || donor_type!=st) {
              donor_identity = id;
              donor_type = st;
              donor_timescale = t;
           }
        }
     }
}


void double_star::set_donor_timescale(const real t, const stellar_type st,
                                      const int id) {

     if (bin_type!=Merged && bin_type!=Disrupted) {
        if (st!=NAS) {
           if (donor_identity==id) {
              donor_type = st;
              donor_timescale = t;
           }
           else if(donor_timescale<=0) {
              donor_identity = id;
              donor_type = st;
              donor_timescale = t;
           }
        }
     }
}


void double_star::spiral_in(star* larger,
                            star* smaller) {
//	Compare total binary energy with the binading energy of the
//	donor's envelope.
//	Magin parameter ALPHA determines efficiency of spiral-in.
//	If two stars are in contact after the common envelope,
//	stars are merged.

    cerr<<"void double_star::spiral_in()"<<endl;
    dump(cerr);
     if (bin_type!=Merged && bin_type!=Disrupted) {

//	increase 'larger' core_mass.
        real mtot_l = larger->get_total_mass();
        real mcl_old = larger->get_core_mass();
        real mel_old = larger->get_envelope_mass();
        real mcore_l = mcl_old;
        real menv_l = mel_old;

        larger->set_core_mass(mcore_l);
        larger->set_envelope_mass(menv_l);
        real mtot_s = smaller->get_total_mass();

        real r_lobe = roche_radius(larger)/semi;
        real a_spi = semi*(mcore_l/mtot_l)/(1. + (2.*menv_l
                   /      (ALPHA*LAMBDA*r_lobe*mtot_s)));
       
        real rl_l = roche_radius(a_spi, mcore_l, mtot_s);
        real rl_s = roche_radius(a_spi, mtot_s, mcore_l);

        if (larger->get_core_radius()>=rl_l 	||
           smaller->get_radius()>=rl_s) 	{
           real da = semi*(1 - r_lobe) - smaller->get_effective_radius();
           real semi_fr = 1-da/semi;
           real mass_lost = (1 - semi_fr)
                          / (1./mtot_l 
                          + 2*semi_fr/(mtot_s*ALPHA*LAMBDA*r_lobe));

           if (mass_lost>=menv_l) {
              mass_lost = 0.99*menv_l;
           }

//		Most elegant solution is exceptions.
//		But my compiler does not allow me! @#&#^@*
/*         try {
 *           larger->reduce_donor_mass(md_dot);
 *         }
 *         catch (star * element) {
 *             larger = element;
 *         }
 */
//			So...
//		Pre- `class star' solution
//           star * old_l = larger;
//           larger->reduce_donor_mass(mass_lost);
//           if (old_l==primary) larger=primary;
//           else larger=secondary;

           larger->set_core_mass(mcl_old);
           larger->set_envelope_mass(mel_old);
//           larger->reduce_donor_mass(mass_lost);
         larger = larger->reduce_donor_mass(mass_lost);
/*
 *       try {
 *          larger->reduce_donor_mass(mass_lost);
 *       }
 *       catch (star* element) {
 *          larger = element;
 *       }
 */

            
           if (mtot_l>mtot_s) 
              merge_elements(larger, smaller);
           else 
              merge_elements(smaller, larger);

        } 
        else {
           semi = a_spi;
        larger = larger->reduce_donor_mass(menv_l);


           smaller->add_mass_to_accretor(menv_l, SPI_TIME);
           set_donor_timescale(smaller);
           determine_minimal_timestep();  
        }
     }
//cerr<<"leave spiral_in: semi = "<<semi<<endl;
}

void double_star::merge_elements(star* consumer,
                                 star* dinner) {

//cerr<<"void double_star::merge_elements()"<<endl;
//cerr<<"bin: "<<identity<<endl;

        bin_type = Merged;
#if 0
//		Merging is performed somwhere in KIRA
//		Not here anymore!  done: 15-08-1996
        dump("merge.dat");
        consumer->set_velocity(velocity); 
        dinner->set_velocity(velocity); 
        consumer->merge_elements(dinner);
        semi = 0;
#endif
     }

void double_star::perform_wind_loss(const real dt) {

        get_primary()->stellar_wind(dt);
        get_secondary()->stellar_wind(dt);

}

void double_star::angular_momentum_loss(const real dt) {
//	Lose angular momentum.
//	First due to magnetic breaking than by gravitational 
//	wave radiation.

//cerr<<"semi="<<semi;

        if (bin_type != Merged && bin_type != Disrupted) 
              magnetic_stellar_wind(dt);

//cerr<<"-->"<<semi<<endl;

     }

void double_star::magnetic_stellar_wind(const real dt) {
//	Lose angular momentum due to magnetic breaking.
//	Rappaport, Verbunt and Joss 1983.
//
//	Magnetic breaking takes place if the binary eccentricity
//	becomes smaller than COROTATION_E_LIMIT due to tidal circularization.

//cerr<<"void double_star::magnetic_stellar_wind(dt="<<dt<<)"<<endl;

     real c_mb = -3.8e-30*M_SUN*G/R_SUN;
     real J_mb_p = 0;
     if (get_primary()->low_mass_star() && get_primary()->get_total_mass()>=M_MSW &&
        !get_primary()->remnant() && eccentricity<=COROTATION_E_LIMIT)
        J_mb_p = c_mb * pow(get_total_mass(), 2)
                    * pow(get_primary()->get_effective_radius(), GAMMA_mb)
                    / (get_secondary()->get_total_mass()*pow(semi, 5));
     real J_mb_s = 0;
     if (get_secondary()->low_mass_star() && get_secondary()->get_total_mass()>=M_MSW &&
        !get_secondary()->remnant() && eccentricity<=COROTATION_E_LIMIT)
        J_mb_s = c_mb * pow(get_total_mass(), 2)
                    * pow(get_secondary()->get_effective_radius(), GAMMA_mb)
                    / (get_primary()->get_total_mass()*pow(semi, 5));

     real a_dot = 2*(J_mb_p + J_mb_s)*dt*M_year;
     real semi_new = semi*(1 + a_dot);

     if (semi_new<=0) {
        if (get_primary()->get_radius()>get_secondary()->get_radius())
           merge_elements(get_primary(), get_secondary());
        else
           merge_elements(get_secondary(), get_primary());
     }
     else if(semi_new <= get_primary()->get_core_radius() + 
                         get_secondary()->get_core_radius() ) {
        semi=semi_new;
        if (get_primary()->get_radius()>get_secondary()->get_radius())
           spiral_in(get_secondary(), get_primary());
        else
           spiral_in(get_primary(), get_secondary());
     }
     else { 
        semi=semi_new;
        gravrad(dt);
     }
}

real double_star::de_dt_gwr(const real e) {
//	Infenitesimal eccentricity variation due to the radiation
//	of gravitational waves.
//	Peters. P.C., 1964, {\em Phys. Rev.}, {\bf 136}, 1224.

//cerr<<"real double_star::de_dt_gwr(const real e="<<e<<")";

        real de_dt = (-305/15)*e*(1 + (121/304)*e*e)/pow(1-e*e, 2.5);
        de_dt *= get_primary()->get_total_mass()*get_secondary()->get_total_mass()
              *  get_total_mass()/pow(semi, 4.);

//cerr<<" = "<<de_dt<<endl;

        return de_dt;
     }

void double_star::gravrad(const real dt) {
//	Equations from:
//	Peters P.C., 1964, Phys. Ref., 136, 4B, B1224.

     real G3M3_C5R4 = pow(G*M_SUN, 3.)
                    / (pow(C_light, 5.)*pow(R_SUN, 4.));

//	Only appicable to small separation.
     if(semi/sqrt(get_total_mass())<=6.) {
        real e_new = eccentricity 
                   + G3M3_C5R4*de_dt_gwr(eccentricity)*dt*M_year;

        real a_new, c0;
        if (e_new>0 && eccentricity>0) {
           c0 = (semi*(1-pow(eccentricity,2))/pow(eccentricity, 12./19.))
                   / pow(1 + 121*pow(eccentricity, 2.)/304, 870./2299.);
       
           a_new = (c0*pow(e_new, 12./19.)/(1-pow(e_new, 2.)))
                   * pow(1 + 121*pow(e_new, 2.)/304, 870./2299.);
        }
        else {
           e_new = 0; 		// Circularize
           semi *= (1-eccentricity*eccentricity);
           real G3M3_C5 = pow(G*M_SUN, 3.)/pow(C_light, 5.);
           c0 = G3M3_C5*128
                       * get_primary()->get_total_mass()
                       * get_secondary()->get_total_mass()
                       * get_total_mass()/5.;
           c0 = pow(semi*R_SUN,4.) - c0*dt*M_year;
           if(c0<0) c0 = 0;
           a_new = pow(c0, 0.25)/R_SUN;
/*
//		Peters method.
           c0 = (-64./5.)*G3M3_C5R4
              * get_primary()->get_total_mass()*get_secondary()->get_total_mass()
              * get_total_mass()/(pow(semi, 3.)*pow(1-eccentricity, 7./2.));
           c0 *= (1 + 73*pow(eccentricity, 2.)/24
                   + 37*pow(eccentricity, 4.)/96);
           a_new = semi + c0*dt*M_year;
*/
        }
//cerr<<" e_new : "<<eccentricity<<" => "<<e_new <<endl;
//cerr<<" a_new : "<<semi <<" => "<<a_new<<endl;
    
        if (a_new<=0) {
           if (get_primary()->get_total_mass()>get_secondary()->get_total_mass())
              merge_elements(get_primary(), get_secondary());
           else
              merge_elements(get_secondary(), get_primary());
        }
        else if(a_new <= get_primary()->get_core_radius() +
                         get_secondary()->get_core_radius() ) {
           semi=a_new;
           eccentricity = e_new;
           if (get_primary()->get_radius()>get_secondary()->get_radius())
              spiral_in(get_secondary(), get_primary());
           else
              spiral_in(get_primary(), get_secondary());
        }
        else {
           semi=a_new;
           eccentricity = e_new;
        }
     }
}

real double_star::mdot_according_to_roche_radius_change(star* donor,
                                                   star* accretor) {
//	Mass lost from donor due to loss of angular momentum by
//	magnetic braking and gravitational wave-radiation.
//	mass-transfer timescale becomes comparable to
//	themal time-scale.

//cerr<<"real double_star::mdot_according_to_roche_radius_change() = ";
        real mdot = 0;
     if (bin_type != Merged && bin_type != Disrupted) {

        real r_don = donor->get_effective_radius();
        real m_don = donor->get_total_mass();
        real m_acc = accretor->get_total_mass();

        real zeta_lobe = zeta(donor, accretor);
        real J_mb = 0;
        if (donor->get_total_mass()>=M_MSW && !donor->remnant()) {
           real c_mb = -3.8e-30*M_SUN*G/R_SUN;
           J_mb = M_year*c_mb * pow(m_don+m_acc, 2)*pow(r_don, GAMMA_mb)
                / (m_acc*pow(semi, 5)); 
        }

        real J_gr=0;
        if (semi/sqrt(get_total_mass())<=6.) {
           real c_gr = -32*pow(M_SUN*G, 3)/(5*pow(C_light, 5)*pow(R_SUN, 4));
           J_gr = M_year*c_gr*m_don*m_acc*(m_don+m_acc)/pow(semi, 4);
        }

//		On the nomination for checking!
        real drdot_dr=0;	//No stellar size variation.
        mdot = m_don*abs((2*(J_mb+J_gr) - drdot_dr)
             /           (zeta_lobe + 2*(5./6. - m_don/m_acc)));
     }
      
//cerr<<mdot<<endl;
     return mdot;
}

void double_star::calculate_velocities() {
//	Compute radial orbital velocities a binary components.
//	neglecting the eccenticity.

     if(bin_type != Merged && bin_type != Disrupted) {
        real mu = get_primary()->get_total_mass()/get_total_mass();
        real mean_r = semi*(1+0.5*eccentricity*eccentricity);
        real v_rel = sqrt(G*get_total_mass()*M_SUN
                   *      ((2/(mean_r*R_SUN)) - (1/(semi*R_SUN))));
        v_rel /= Km_s;
        get_primary()->set_velocity((1-mu)*v_rel);
        get_secondary()->set_velocity(mu*v_rel);
     }
}

void double_star::common_envelope(real dt) {
//	Alternative of the double_star::contact();
//	No mass transfer occurs during stable case.
    cerr << "common envelope "<<endl;
    dump(cerr);
     if (!stable()) {
        if (get_primary()->get_radius()>get_secondary()->get_radius())
           spiral_in(get_primary(), get_secondary());
        else
           spiral_in(get_secondary(), get_primary());
     }
     else {
//		Now thermal unstable stars may spiral-in.
//		Generally condition causes main-sequence stars to
//		
//cerr<<"ce: z"<<get_primary()->zeta_adiabatic()<<" "<<get_secondary()->zeta_adiabatic()<<endl;
        if (get_primary()->zeta_adiabatic()<=0.5	||
            get_secondary()->zeta_adiabatic()<=0.5) 	{
           if (get_primary()->get_radius()>get_secondary()->get_radius())
              spiral_in(get_primary(), get_secondary());
           else
              spiral_in(get_secondary(), get_primary());
        }
        else {
//		Spiral in slowly or if smaller is allowed to accrete
//		Spiral out slowly.
//		Presenlt leave system unchanged, finally spiral-in will
//		occur.
//		Results in flip-flop (spiral-in followed by spiral out.
//		Causes stars with infinite lifetime.
//        if (primary->get_radius()>secondary->get_radius()) {
//           if (ready_for_mass_transfer(secondary))
//              contact(secondary, primary, dt);
//        }
//        else {
//           if (ready_for_mass_transfer(primary)) 
//              contact(primary, secondary, dt);
//        }
        }
     }
cerr<<"Leave common envelope"<<endl;    
}



void double_star::refresh_memory() {

cerr<<"void double_star::refresh_memory()"<<endl;

     get_primary()->refresh_memory();
     get_secondary()->refresh_memory();

     previous.binary_age   = binary_age;
     previous.semi         = semi;
     previous.eccentricity = eccentricity;
cerr<<"leave refresh_mory()"<<endl;

}

void double_star::recall_memory() {
//cerr<<"void double_star::recall_memory()"<<endl;

     get_primary()->recall_memory();
     get_secondary()->recall_memory();

     binary_age   = previous.binary_age;
     semi         = previous.semi;
     eccentricity = previous.eccentricity;
}

real double_star::zeta(star * donor, 
                       star * accretor) {
//	Find zeta Roche-lobe.

//cerr<<"double_star::zeta: a0, a1:"<<previous.semi<<" "<<semi<<endl;
     if (bin_type!=Merged && bin_type!=Disrupted) {
        real md_dot, ma_dot;
        if (!donor->remnant())
           md_dot = ma_dot = 0.001*donor->get_total_mass();
        else
           md_dot = ma_dot = 0.0001*donor->get_core_mass();

//		Accretion from MS to remnant conservative.
        if (accretor->remnant() && donor->get_element_type()!=Main_Sequence)
           ma_dot *= 0.1;

        if (md_dot<=0)
           md_dot = ma_dot = .01;

        real M_old = get_total_mass();
        real old_donor_mass = donor->get_total_mass();
        real old_accretor_mass = accretor->get_total_mass();
        real q_old = old_accretor_mass/old_donor_mass;

        real M_new = get_total_mass() - md_dot +  ma_dot;
        real new_donor_mass = donor->get_total_mass() - md_dot;
        real new_accretor_mass = accretor->get_total_mass() + ma_dot;
        real q_new = new_accretor_mass/new_donor_mass;

        real a_fr, new_semi;
        if (!accretor->remnant()) {
           a_fr  = pow(old_donor_mass*old_accretor_mass
                 / (new_donor_mass*new_accretor_mass), 2);
           new_semi = semi*pow(M_new/M_old, 2*3 + 1)*a_fr;
        }
        else {
           real alpha = 1. - ma_dot/md_dot;
           a_fr  = (new_donor_mass/old_donor_mass)
                 * pow(new_accretor_mass/old_accretor_mass, 1./(1.-alpha));
           new_semi = semi*(M_old/M_new)/pow(a_fr, 2);
        }

/*
//              Intermezzo.
     real a_dot=0, J_mb = 0, J_gr=0;
     if(accretor->remnant() && donor->get_element_type()==Main_Sequence) {
     if (donor->get_total_mass()>=M_MSW && !donor->remnant()) {
       real c_mb = -3.8e-30*M_SUN*G/R_SUN;
       J_mb = M_year*c_mb * pow(M_old, 2)
            * pow(donor->get_effective_radius(), GAMMA_mb)
                 / (old_accretor_mass*pow(semi, 5));
     }

     if (semi/sqrt(get_total_mass())<=6.) {
        real c_gr = -32*pow(M_SUN*G, 3)/(5*pow(C_light, 5)*pow(R_SUN, 4));
        J_gr = M_year*c_gr*old_donor_mass*old_accretor_mass*(M_old)/pow(semi, 4);
     }
     a_dot = semi*2*(J_mb + J_gr);
     }
    a_dot = 0;
    new_semi += a_dot;
*/

//		NEW improved(?) method:
//		Seems to work fine (see model S)
        real mdo = donor->get_previous_total_mass();
        real md1 = donor->get_total_mass();
        real mao = accretor->get_previous_total_mass();
        real ma1 = accretor->get_total_mass();
        real mdd = (md1-mdo)/mdo;
        mdd = min(mdd, -0.00001);
        real mda = (ma1-mao)/mao;
        mda = max(mda, 0.0);
//cerr<<"mda: "<<mdo<<" "<<md1<<" "<<mao<<" "<<ma1<<" "<<mdd<<" "<<mda<<endl;
        real rl_old =  roche_radius(previous.semi, mdo, mao);
        real rl_new = roche_radius(donor);
        real drldot_rl = (rl_new-rl_old)/rl_old;
//        real drldot_rl = (semi-previous.semi)/semi
//                       + 2*mdd + 2*mda
//                       - (md1+ma1-(mdo+mao))/(mdo+mao) 
//                       - 2*mdd*(5./6. - mdo/mao);
//        cerr<<" "<<drldot_rl<<" "<<((md1+ma1-(mdo+mao))/(mdo+mao))<<endl;
//        real zeta = log(abs(drldot_rl))/mdd;
        real zeta = drldot_rl/(mdd);
//cerr<<"zeta: "<<zeta<<endl;
	return zeta;
   }

/*

//		SKIP this.... since improved(?) method is put on.
//	Numerically stable but conservative.
        real rl = roche_radius(donor);
//      Eggleton should give better results.
//        real rl_d = 0.52*new_semi
//                  * pow(new_donor_mass/(M_new),0.44);
        real rl_d = roche_radius(new_semi, new_donor_mass, new_accretor_mass);

        real d_lnr = (rl - rl_d)/rl;
        real d_lnm = (donor->get_total_mass() - new_donor_mass)
                   /  donor->get_total_mass();

//cerr<<"zeta (old): "<<d_lnr/d_lnm<<endl;
//cerr<<"zeta: "<<rl<<" "<<rl_d<<" "<<donor->get_total_mass()<<" "<<new_donor_mass<<" "<<semi<<" "<<new_semi<<d_lnr<<" "<<d_lnm<<endl;
        return d_lnr/d_lnm;
     }
    
*/
     return 1;
   }

//	Gives the linker a headacke.
void double_star::enhance_cluster_profile(cluster_profile& c_prof) {

      double_profile binary;
      //make_profile(identity, initial.start_time, binary, initial);
      binary.enhance_double_profile(this);
      c_prof.enhance_cluster_profile(binary);      
   }
